package hep.wired.corbaheprep;

import java.io.IOException;
import java.util.*;
import org.omg.CORBA.Any;
import org.omg.CORBA.SystemException;

import hep.graphics.heprep.HepRep;
import hep.graphics.heprep.HepRepAttValue;
import hep.graphics.heprep.HepRepInstanceTree;
import hep.graphics.heprep.HepRepProvider;
import hep.graphics.heprep.HepRepTreeID;
import hep.graphics.heprep.HepRepType;
import hep.graphics.heprep.HepRepTypeTree;
import hep.graphics.heprep.util.HepRepUtil;
import hep.wired.hepeventserver.idl.HepEventServer;
import org.freehep.application.Application;
import org.freehep.application.studio.Studio;
import org.freehep.record.source.AbstractRecordSource;
import org.freehep.record.source.DefaultRecordTag;
import org.freehep.record.source.NoSuchRecordException;
import org.freehep.record.source.RecordSource;
import org.freehep.record.source.RecordTag;


/**
 * Basic implementation of {@link RecordSource} wrapped around CORBA HepRep server.
 * Supports sequential access only, assumes the next event can be loaded by
 * <tt>HepEventServer.setEvent("next")</tt>. Designed to be used as a base class for
 * implementing wrappers for specific CORBA servers.
 *
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public class CORBARecordSource extends AbstractRecordSource {
  
  public static enum Move {
    CURRENT, 
    NEXT, 
    PREVIOUS, 
    INDEX, 
    TAG, 
    SHIFT, 
    REWIND;
  };

// -- Private parts : ----------------------------------------------------------
  
  static final protected long UNKNOWN = -2L;

  protected HepEventServer _server; // CORBA server
  
  protected long _size; // source size if known, UNKNOWN otherwise
  protected EnumSet<Move> _supports; // flags indicating support for positioning operations
  protected boolean _isHepRep2; // flag indicating that data is in HepRep2 format
  
  protected long _index; // current index, -1 if before first event, UNKNOWN if unknown
  protected RecordTag _tag; // current tag, null if there is no current event
  protected HepRep _heprep;  // current event, null if none
  
  protected boolean _hasNext;
  protected boolean _hasPrevious;

  private HepRepProvider _converter;

  final private Map<String,HepRepInstanceTree> _cachedGeometry = new HashMap<String,HepRepInstanceTree>();
  final private Map<String,HepRepTypeTree> _cachedGeometryTypes = new HashMap<String,HepRepTypeTree>();
  private static List<String> _cacheList = Collections.emptyList();


// -- Construction and initialization : ----------------------------------------

  /** Names of trees to be cached. */
  static {
    String cacheable = System.getProperty("hep.wired.corbaheprep.cacheable");
    if (cacheable != null) {
      _cacheList = Arrays.asList(cacheable.split("\\s"));
    }
  }

  public CORBARecordSource(String name, HepEventServer server, Properties serverInfo) {
    
    super(name +" ("+ serverInfo.get("servername") +")");
    _server = server;
    
    _size = UNKNOWN;
    determineSize();
    
    _supports = EnumSet.of(Move.NEXT);
    
    _isHepRep2 = "heprep2".equalsIgnoreCase(serverInfo.getProperty("dataformat"));

    _index = -1L;
    _tag = null;
    _heprep = null;
    _hasNext = true;
    _hasPrevious = true;

    _converter = null;
  }
  
  protected final void determineSize() {
    try {
      long size = _server.getNumberOfEvents();
      if (size > 0L) {
        _size = size;
      } else {
        _size = UNKNOWN;
      }
    } catch (Exception x) {
      _size = UNKNOWN;
    }
  }

// -- Source parameters : ------------------------------------------------------

  /** Returns the <tt>Class</tt> object that is guaranteed to be a superclass of all records in this source. */
  public Class<?> getRecordClass() {
    return HepRep.class;
  }

  /**
   * Returns the number of records in this source, after updating it from the CORBA server.
   * 
   * @throws UnsupportedOperationException  if the number of records is unknown.
   */
  public long size() {
    determineSize();
    if (_size == UNKNOWN) throw new UnsupportedOperationException();
    return _size;
  }

  /**
   * Returns the estimated number of records in this source.
   * Uses previously determined number of records, does not communicate with CORBA server.
   * 
   * @throws UnsupportedOperationException  if the number of records is unknown.
   */
  public long getEstimatedSize() {
    if (_size == UNKNOWN) throw new UnsupportedOperationException();
    return _size;
  }

  
// -- Current record data : ----------------------------------------------------
  
  /**
   * Returns the index of the current record, or <tt>-1</tt> if this source is positioned before the first record.
   */
  public long getCurrentIndex() {
    if (!supportsIndex()) throw new UnsupportedOperationException();
    if (_index == UNKNOWN) throw new IllegalStateException();
    return _index;
  }
  
  /**
   * Returns the tag of the current record.
   * 
   * @throws IllegalStateException  if there is no current record.
   */
  public RecordTag getCurrentTag() {
    if (!supportsTag()) throw new UnsupportedOperationException();
    if (_index == -1L || _tag == null) throw new IllegalStateException();
    return _tag;
  }

  /**
   * Returns the current record.
   * 
   * @throws IllegalStateException  if there is no current record.
   * @throws IOException  if retrieving the current record fails for any reason.
   */
  public HepRep getCurrentRecord() throws IOException {
    if (_index == -1L) throw new IllegalStateException();
    if (_heprep == null) _heprep = getHepRep();
    return _heprep;
  }

  /** Releases any resources associated with the current record. */
  public void releaseRecord() {
    _heprep = null;
  }
  
  
// -- Positioning : ------------------------------------------------------------
  
  /** Returns <tt>true</tt> if this source supports reloading current record. */
  public boolean supportsCurrent() {
    return _supports.contains(Move.CURRENT);
  }
  
  /** Returns <tt>true</tt> if this source supports loading next record. */
  public boolean supportsNext() {
    return _supports.contains(Move.NEXT);
  }
  
  /** Returns <tt>true</tt> if this source supports loading previous record. */
  public boolean supportsPrevious() {
    return _supports.contains(Move.PREVIOUS);
  }
  
  /** Returns <tt>true</tt> if this source supports selecting records by index. */
  public boolean supportsIndex() {
    return _supports.contains(Move.INDEX);
  }
  
  /** Returns <tt>true</tt> if this source supports selecting records by tag. */
  public boolean supportsTag() {
    return _supports.contains(Move.TAG);
  }
  
  /** Returns <tt>true</tt> if this source supports selecting records by offset with respect to the current record. */
  public boolean supportsShift() {
    return _supports.contains(Move.SHIFT);
  }

  /** Returns <tt>true</tt> if this source can reload the current record. */
  public boolean hasCurrent() {
    return supportsCurrent() && (_index >= 0L || _tag != null);
  }

  /** Returns <tt>true</tt> if this source can load the next record. */
  public boolean hasNext() {
    return supportsNext() && _hasNext && (_index < _size-1 || _size == UNKNOWN);
  }

  /** Returns <tt>true</tt> if this source can load the previous record. */
  public boolean hasPrevious() {
    return supportsPrevious() && _hasPrevious && _index != -1L;
  }

  /** Returns <tt>true</tt> if this source has a record with the specified index. */
  public boolean hasIndex(long index) {
    return supportsIndex() && -1L < index && (index < _size || _size == UNKNOWN);
  }

  /** Returns <tt>true</tt> if this source can shift by <tt>numberOfRecords</tt> records. */
  public boolean hasShift(long numberOfRecords) {
    if (!supportsShift()) return false;
    long index = _index + numberOfRecords;
    return -1L < index && (index < _size || _size == UNKNOWN);
  }

  /**
   * Reloads the current record.
   *
   * @throws NoSuchRecordException if there is no current record.
   * @throws IOException  if reloading of the current record fails.
   * @throws UnsupportedOperationException if this source does not support reloading of the current record.
   */
  public void current() throws IOException, NoSuchRecordException {
    if (!supportsCurrent()) throw new UnsupportedOperationException();
    if (_index < 0L && _tag == null) throw new NoSuchRecordException();
    _heprep = null;
  }

  /**
   * Loads the next record.
   * 
   * @throws NoSuchRecordException if there is no current record.
   * @throws IOException if loading of the next record fails.
   * @throws UnsupportedOperationException if this source does not support loading the next record.
   */
  public void next() throws IOException, NoSuchRecordException {
    if (!supportsNext()) throw new UnsupportedOperationException();
    try {
      positionServer(Move.NEXT, 0L, null);
      if (_index != UNKNOWN) _index++;
      setCurrentTag();
    } catch (NoSuchRecordException x) {
      _hasNext = false;
      throw x;
    }
  }

  /**
   * Loads the previous record.
   * 
   * @throws NoSuchRecordException if there is no previous record.
   * @throws IOException if loading of the previous record fails.
   * @throws UnsupportedOperationException if this source does not support loading the previous record.
   */
  public void previous() throws IOException, NoSuchRecordException {
    if (!supportsPrevious()) throw new UnsupportedOperationException();
    try {
      positionServer(Move.PREVIOUS, 0L, null);
      if (_index != UNKNOWN) {
        if (--_index == UNKNOWN) setCurrentIndex();
      }
      setCurrentTag();
    } catch (NoSuchRecordException x) {
      _hasPrevious = false;
      throw x;
    }
  }

  /**
   * Loads the record specified by the index.
   * 
   * @throws NoSuchRecordException if this source does not have a record with the specified index.
   * @throws IOException if loading of the requested record fails.
   * @throws UnsupportedOperationException if this source does not support access by index.
   */
  public void jump(long index) throws IOException, NoSuchRecordException {
    if (!supportsIndex()) throw new UnsupportedOperationException();
    positionServer(Move.INDEX, index, null);
    _index = index;
    setCurrentTag();
  }

  /**
   * Loads the record specified by the tag.
   * 
   * @throws NoSuchRecordException if this source does not have a record with the specified tag.
   * @throws IOException if loading of the requested record fails.
   * @throws UnsupportedOperationException if this source does not support access by tag.
   */
  public void jump(RecordTag tag) throws IOException, NoSuchRecordException {
    if (!supportsTag()) throw new UnsupportedOperationException();
    positionServer(Move.TAG, 0L, tag);
    _tag = tag;
    setCurrentIndex();
  }

  /**
   * Loads the record specified by the offset with respect to the current cursor position. 
   * The offset can be negative or positive. Skipping is guaranteed to be at least as fast as
   * <tt>numberOfRecords</tt> consecutive calls to <tt>next()</tt> or <tt>previous()</tt>, and
   * will be faster if this source implementation permits it.
   * 
   * @throws NoSuchRecordException if this source does not have a record with the specified offset.
   * @throws IOException if loading of the requested record fails.
   * @throws UnsupportedOperationException if this source does not support access by offset.
   */
  public void shift(long numberOfRecords) throws IOException, NoSuchRecordException {
    if (!supportsShift()) throw new UnsupportedOperationException();
    positionServer(Move.SHIFT, numberOfRecords, null);
    if (_index != UNKNOWN) {
      _index += numberOfRecords;
      if (_index < -1L) setCurrentIndex();
    }
    setCurrentTag();
  }
 
  /** 
   * Returns <tt>true</tt> if this source supports rewind operation.
   */
  public boolean supportsRewind() {
    return _supports.contains(Move.REWIND);
  }
  
  /** 
   * Returns <tt>true</tt> if this source in its current state can be rewound.
   */
  public boolean hasRewind() {
    return supportsRewind() && _index != -1L;
  }
  
  /**
   * Positions the cursor of this source before the first record.
   */
  public void rewind() throws IOException {
    if (!supportsRewind()) throw new UnsupportedOperationException();
    try {
      positionServer(Move.REWIND, 0L, null);
      _index = -1L;
      _tag = null;
    } catch (NoSuchRecordException x) {
      throw new UnsupportedOperationException(x);
    }
  }
  
// -- Closing the source : -----------------------------------------------------

  /**
   * Close the record source and release any associated resources.
   * 
   * @throws IOException  if closing fails.
   */
  public void close() throws IOException {
    releaseRecord();
    if (_server != null) {
      try {
        _server.release();
      } catch (SystemException x) {}
    }
    _server = null;
  }

  
// -- Methods that can be customized by subclusses : ---------------------------
  
  /**
   * Returns the command that needs to be send to the CORBA server to position it for reading the specified record.
   */
  protected String getServerCommand(Move move, long par, RecordTag tag) {
    if (move != Move.NEXT) throw new UnsupportedOperationException();
    return "next";
  }
  
  /** Sets <tt>_tag</tt> to the tag of the record for which the CORBA server is currently positioned. */
  protected void setCurrentTag() {
    try {
      _tag = new DefaultRecordTag(_server.getEventTitle());
    } catch (Exception x) {
      _tag = null;
    }
  }
  
  /** Sets <tt>_index</tt> to the index of the record for which the CORBA server is currently positioned. */
  protected void setCurrentIndex() {
    _index = UNKNOWN;
  }
  
  protected void positionServer(Move move, long par, RecordTag tag) throws IOException, NoSuchRecordException {
    String command = getServerCommand(move, par, tag);
    _heprep = null;
    boolean ok = false;
    try {
      sendCommand(command);
      ok = true;
    } finally {
      _hasNext = true;
      _hasPrevious = true;
      if (!ok) restorePosition();
    }
  }
  
  protected void restorePosition() {
    try {
      if (supportsTag() && _tag != null) {
        sendCommand(getServerCommand(Move.TAG, 0L, _tag));
      } else if (supportsIndex() && _index >= 0L) {
        sendCommand(getServerCommand(Move.INDEX, _index, null));
      }
      return;
    } catch (NoSuchRecordException x) {
    } catch (IOException x) {}
    _index = UNKNOWN;
    _tag = null;
  }
  
  protected void sendCommand(String command) throws NoSuchRecordException, IOException {
    try {
      _server.setEvent(command);
    } catch (SystemException x) {
      throw new IOException("CORBA exception processing command: "+ command, x);
    } catch (Exception x) {
      throw new NoSuchRecordException(x);
    }
  }
  
  protected HepRep getHepRep() throws IOException {
    
    HepRep heprep = null;

    try {

      Any hepRepAny = null;
      try {
        hepRepAny = _server.getEventData("HepRep");
      } catch (SystemException x) {
        throw new IOException(x);
      }
      org.omg.CORBA.Object anyObj = hepRepAny.extract_Object();


      if (_isHepRep2) {
        // HepRep 2
        hep.graphics.heprep.corba.idl.HepRep heprepObject = hep.graphics.heprep.corba.idl.HepRepHelper.narrow(anyObj);
        heprep = new hep.graphics.heprep.corba.HepRepAdapter(heprepObject);

        // Get the tops of the event
        String eventName = _server.getEventTitle();
        List tops = getInstanceTreeTops(heprep, heprep.getInstanceTreeTop(eventName, "1.0"));

        for (Iterator i = tops.iterator(); i.hasNext();) {
          // retrieve the type tree
          HepRepInstanceTree instanceTreeTop = (HepRepInstanceTree) i.next();
          HepRepTreeID id = instanceTreeTop.getTypeTree();

          HepRepTypeTree typeTree = _cachedGeometryTypes.get(id.getName());
          if (typeTree == null) {
            typeTree = heprep.getTypeTree(id.getName(), id.getVersion());
            if (_cacheList.contains(id.getName())) {
              _cachedGeometryTypes.put(id.getName(), typeTree);
            }
          } else {
            heprep.removeTypeTree(heprep.getTypeTree(id.getName(), id.getVersion()));
            heprep.addTypeTree(typeTree);
          }

          HepRepInstanceTree instanceTree = _cachedGeometry.get(id.getName());
          if (instanceTree == null) {
            // retrieve the instance tree
            instanceTree = heprep.getInstances(instanceTreeTop.getName(), instanceTreeTop.getVersion(), getVisibleTypes(typeTree));
            if (_cacheList.contains(id.getName())) {
              _cachedGeometry.put(id.getName(), instanceTree);
            }
          }
          heprep.removeInstanceTree(instanceTreeTop);
          heprep.addInstanceTree(instanceTree);
        }
      } else {
        // HepRep 1
        hep.graphics.heprep1.corba.idl.HepRep heprep1Object = hep.graphics.heprep1.corba.idl.HepRepHelper.narrow(anyObj);
        hep.graphics.heprep1.HepRep heprep1 = new hep.graphics.heprep1.corba.HepRepAdapter(heprep1Object);

        // Looking for Type Info
        for (Enumeration e = heprep1.getTypeInfo(); e.hasMoreElements();) {
          hep.graphics.heprep1.HepRepTypeInfo typeInfo = (hep.graphics.heprep1.HepRepTypeInfo) e.nextElement();
          hep.graphics.heprep1.HepRepAttValue attValue = typeInfo.getAttValue("visibility");
          boolean visible = attValue == null ? true : attValue.getBoolean();

          if (visible) {
            // fill up HepRep with its known types, automatically
            // added.
                        /* hep.graphics.heprep1.HepRepType type = */ heprep1.getRepresentablesUncut(typeInfo.getName(),
                    typeInfo.getVersion());
          }
          // FIXME add others here for deferred loading
        }

        // look for converter
        if (_converter == null) {
          _converter = HepRepUtil.getHepRepProvider(((Studio) Application.getApplication()).getLookup(), heprep1);
        }

        if (_converter == null) {
          throw new IOException("Cannot find HepRep1 to HepRep2 converter");
        }
        heprep = _converter.convert(heprep1);
      }
    } catch (SystemException x) {
      throw new IOException(x);
    }
    return heprep;
  }
  

// -- Local methods : ----------------------------------------------------------

  // FIXME, maybe part of HepRepUtils ?
  private List/* <HepRepTreeID> */ getInstanceTreeTops(HepRep heprep, HepRepInstanceTree top) {
    List tops = new ArrayList();
    tops.add(top);

    // Loop to get all referred instance tree tops
    // retrieve all instance tree tops
    for (Iterator i = top.getInstanceTreeList().iterator(); i.hasNext();) {
      HepRepTreeID id = (HepRepTreeID) i.next();
      top = heprep.getInstanceTreeTop(id.getName(), id.getVersion());
      tops.addAll(getInstanceTreeTops(heprep, top));
    }
    return tops;
  }

  // FIXME, maybe part of HepRepUtils ?
  private String[] getVisibleTypes(HepRepTypeTree typeTree) {
    List visibles = getVisibleTypes(typeTree.getTypeList());
    return (String[]) visibles.toArray(new String[visibles.size()]);
  }

  // FIXME, maybe part of HepRepUtils ?
  private List/* <String> */ getVisibleTypes(List/* <HepRepType> */ types) {
    List visibles = new ArrayList();
    for (Iterator i = types.iterator(); i.hasNext();) {
      HepRepType type = (HepRepType) i.next();
      HepRepAttValue attValue = type.getAttValue("visibility");
      boolean visible = attValue == null ? true : attValue.getBoolean();
      if (visible) {
        visibles.add(type.getFullName());
      }
      visibles.addAll(getVisibleTypes(type.getTypeList()));
    }
    return visibles;
  }
  
// -----------------------------------------------------------------------------
}
